@@ -19,7 +19,8 @@ from .errors import PdfParseError, log
 from .tokens import PdfTokens
 from .objects import PdfDict, PdfArray, PdfName, PdfObject, PdfIndirect
 from .uncompress import uncompress
-from .py23_diffs import convert_load, iteritems
+from . import crypt
+from .py23_diffs import convert_load, convert_store, iteritems
 
 
 class PdfReader(PdfDict):
@@ -265,8 +266,17 @@ class PdfReader(PdfDict):
             for key in new:
                 self.loadindirect(key)
 
+    def decrypt_all(self):
+        self.read_all()
+
+        if self.crypt_filters is not None:
+            crypt.decrypt_objects(
+                self.indirect_objects.values(), self.stream_crypt_filter,
+                self.crypt_filters)
+
     def uncompress(self):
         self.read_all()
+
         uncompress(self.indirect_objects.values())
 
     def load_stream_objects(self, object_streams):
@@ -279,22 +289,26 @@ class PdfReader(PdfDict):
 
         # read objects from stream
         if objs:
+            # Decrypt
+            if self.crypt_filters is not None:
+                crypt.decrypt_objects(
+                    objs, self.stream_crypt_filter, self.crypt_filters)
+
+            # Decompress
             uncompress(objs)
+
             for obj in objs:
                 objsource = PdfTokens(obj.stream, 0, False)
-                snext = objsource.next
-                offsets = {}
+                next = objsource.next
+                offsets = []
                 firstoffset = int(obj.First)
-                num = snext()
-                while num.isdigit():
-                    offset = int(snext())
-                    offsets[int(num)] = firstoffset + offset
-                    num = snext()
-                for num, offset in iteritems(offsets):
+                while objsource.floc < firstoffset:
+                    offsets.append((int(next()), firstoffset + int(next())))
+                for num, offset in offsets:
                     # Read the object, and call special code if it starts
                     # an array or dictionary
                     objsource.floc = offset
-                    sobj = snext()
+                    sobj = next()
                     func = self.special.get(sobj)
                     if func is not None:
                         sobj = func(objsource)
@@ -332,7 +346,6 @@ class PdfReader(PdfDict):
         '''
 
         def readint(s, lengths):
-            lengths = itertools.cycle(lengths)
             offset = 0
             for length in itertools.cycle(lengths):
                 next = offset + length
@@ -354,8 +367,13 @@ class PdfReader(PdfDict):
             source.exception('Expected dict type of /XRef')
         tok = next()
         self.readstream(obj, self.findstream(obj, tok, source), source, True)
+        old_strm = obj.stream
         if not uncompress([obj], True):
             source.exception('Could not decompress Xref stream')
+        stream = obj.stream
+        # Fix for issue #76 -- goofy compressed xref stream
+        # that is NOT ACTUALLY COMPRESSED
+        stream = stream if stream is not old_strm else convert_store(old_strm)
         num_pairs = obj.Index or PdfArray(['0', obj.Size])
         num_pairs = [int(x) for x in num_pairs]
         num_pairs = zip(num_pairs[0::2], num_pairs[1::2])
@@ -363,7 +381,7 @@ class PdfReader(PdfDict):
         if len(entry_sizes) != 3:
             source.exception('Invalid entry size')
         object_streams = defaultdict(list)
-        get = readint(obj.stream, entry_sizes)
+        get = readint(stream, entry_sizes)
         for objnum, size in num_pairs:
             for cnt in range(size):
                 xtype, p1, p2 = islice(get, 3)
@@ -431,7 +449,10 @@ class PdfReader(PdfDict):
         ''' Parse (one of) the cross-reference file section(s)
         '''
         next = source.next
-        tok = next()
+        try:
+            tok = next()
+        except StopIteration:
+            tok = ''
         if tok.isdigit():
             return self.parse_xref_stream(source), True
         elif tok == 'xref':
@@ -450,36 +471,92 @@ class PdfReader(PdfDict):
         typename = PdfName.Type
         kidname = PdfName.Kids
 
-        # PDFs can have arbitrarily nested Pages/Page
-        # dictionary structures.
-        def readnode(node):
-            nodetype = node[typename]
-            if nodetype == pagename:
-                yield node
-            elif nodetype == pagesname:
-                for node in node[kidname]:
-                    for node in readnode(node):
-                        yield node
-            elif nodetype == catalogname:
-                for node in readnode(node[pagesname]):
-                    yield node
-            else:
-                log.error('Expected /Page or /Pages dictionary, got %s' %
-                          repr(node))
         try:
-            return list(readnode(node))
+            result = []
+            stack = [node]
+            append = result.append
+            pop = stack.pop
+            while stack:
+                node = pop()
+                nodetype = node[typename]
+                if nodetype == pagename:
+                    append(node)
+                elif nodetype == pagesname:
+                    stack.extend(reversed(node[kidname]))
+                elif nodetype == catalogname:
+                    stack.append(node[pagesname])
+                else:
+                    log.error('Expected /Page or /Pages dictionary, got %s' %
+                            repr(node))
+            return result
         except (AttributeError, TypeError) as s:
             log.error('Invalid page tree: %s' % s)
             return []
 
-    def __init__(self, fname=None, fdata=None, decompress=False,
-                 disable_gc=True, verbose=True):
+    def _parse_encrypt_info(self, source, password, trailer):
+        """Check password and initialize crypt filters."""
+        # Create and check password key
+        key = crypt.create_key(password, trailer)
 
+        if not crypt.check_user_password(key, trailer):
+            source.warning('User password does not validate')
+
+        # Create default crypt filters
+        private = self.private
+        crypt_filters = self.crypt_filters
+        version = int(trailer.Encrypt.V or 0)
+        if version in (1, 2):
+            crypt_filter = crypt.RC4CryptFilter(key)
+            private.stream_crypt_filter = crypt_filter
+            private.string_crypt_filter = crypt_filter
+        elif version == 4:
+            if PdfName.CF in trailer.Encrypt:
+                for name, params in iteritems(trailer.Encrypt.CF):
+                    if name == PdfName.Identity:
+                        continue
+
+                    cfm = params.CFM
+                    if cfm == PdfName.AESV2:
+                        crypt_filters[name] = crypt.AESCryptFilter(key)
+                    elif cfm == PdfName.V2:
+                        crypt_filters[name] = crypt.RC4CryptFilter(key)
+                    else:
+                        source.warning(
+                            'Unsupported crypt filter: {}, {}'.format(
+                                name, cfm))
+
+            # Read default stream filter
+            if PdfName.StmF in trailer.Encrypt:
+                name = trailer.Encrypt.StmF
+                if name in crypt_filters:
+                    private.stream_crypt_filter = crypt_filters[name]
+                else:
+                    source.warning(
+                        'Invalid crypt filter name in /StmF:'
+                        ' {}'.format(name))
+
+            # Read default string filter
+            if PdfName.StrF in trailer.Encrypt:
+                name = trailer.Encrypt.StrF
+                if name in crypt_filters:
+                    private.string_crypt_filter = crypt_filters[name]
+                else:
+                    source.warning(
+                        'Invalid crypt filter name in /StrF:'
+                        ' {}'.format(name))
+        else:
+            source.warning(
+                'Unsupported Encrypt version: {}'.format(version))
+
+    def __init__(self, fname=None, fdata=None, decompress=False,
+                 decrypt=False, password='', disable_gc=True, verbose=True):
         self.private.verbose = verbose
+
         # Runs a lot faster with GC off.
         disable_gc = disable_gc and gc.isenabled()
         if disable_gc:
             gc.disable()
+
         try:
             if fname is not None:
                 assert fdata is None
@@ -494,8 +571,10 @@ class PdfReader(PdfDict):
                     except IOError:
                         raise PdfParseError('Could not read PDF file %s' %
                                             fname)
-                    fdata = convert_load(fdata)
+
             assert fdata is not None
+            fdata = convert_load(fdata)
+
             if not fdata.startswith('%PDF-'):
                 startloc = fdata.find('%PDF-')
                 if startloc >= 0:
@@ -548,6 +627,23 @@ class PdfReader(PdfDict):
                 xref_list.append((source.obj_offsets, trailer, is_stream))
                 source.floc = int(prev)
 
+            # Handle document encryption
+            private.crypt_filters = None
+            if decrypt and PdfName.Encrypt in trailer:
+                identity_filter = crypt.IdentityCryptFilter()
+                crypt_filters = {
+                    PdfName.Identity: identity_filter
+                }
+                private.crypt_filters = crypt_filters
+                private.stream_crypt_filter = identity_filter
+                private.string_crypt_filter = identity_filter
+
+                if not crypt.HAS_CRYPTO:
+                    raise PdfParseError(
+                        'Install PyCrypto to enable encryption support')
+
+                self._parse_encrypt_info(source, password, trailer)
+
             if is_stream:
                 self.load_stream_objects(trailer.object_streams)
 
@@ -566,6 +662,10 @@ class PdfReader(PdfDict):
                     float(trailer.Version) > float(self.version)):
                 self.private.version = trailer.Version
 
+            if decrypt:
+                self.decrypt_all()
+                trailer.Encrypt = None
+
             if is_stream:
                 self.Root = trailer.Root
                 self.Info = trailer.Info
